@@ -1,16 +1,17 @@
 use crate::module::{
-    Initializer, InstanceSignature, MemoryInitialization, MemoryInitializer, MemoryPlan, Module,
-    ModuleSignature, ModuleType, ModuleUpvar, TableInitializer, TablePlan, TypeTables,
+    AnyfuncIndex, Initializer, InstanceSignature, MemoryInitialization, MemoryInitializer,
+    MemoryPlan, Module, ModuleSignature, ModuleType, ModuleUpvar, TableInitializer, TablePlan,
+    TypeTables,
 };
 use crate::{
     DataIndex, DefinedFuncIndex, ElemIndex, EntityIndex, EntityType, FuncIndex, Global,
     GlobalIndex, GlobalInit, InstanceIndex, InstanceTypeIndex, MemoryIndex, ModuleIndex,
-    ModuleTypeIndex, PrimaryMap, SignatureIndex, TableIndex, Tunables, TypeIndex, WasmError,
-    WasmFuncType, WasmResult,
+    ModuleTypeIndex, PrimaryMap, SignatureIndex, TableIndex, TableInitialization, Tunables,
+    TypeIndex, WasmError, WasmFuncType, WasmResult,
 };
 use cranelift_entity::packed_option::ReservedValue;
 use std::borrow::Cow;
-use std::collections::{hash_map::Entry, HashMap, HashSet};
+use std::collections::{hash_map::Entry, HashMap};
 use std::convert::{TryFrom, TryInto};
 use std::mem;
 use std::path::PathBuf;
@@ -61,18 +62,6 @@ pub struct ModuleTranslation<'data> {
     /// References to the function bodies.
     pub function_body_inputs: PrimaryMap<DefinedFuncIndex, FunctionBodyData<'data>>,
 
-    /// The set of defined functions within this module which are "possibly
-    /// exported" which means that the host can possibly call them. This
-    /// includes functions such as:
-    ///
-    /// * Exported functions
-    /// * Functions in element segments
-    /// * Functions via `ref.func` instructions
-    ///
-    /// This set is used to determine the set of type signatures that need
-    /// trampolines for the host to call into.
-    pub escaped_funcs: HashSet<DefinedFuncIndex>,
-
     /// A list of type signatures which are considered exported from this
     /// module, or those that can possibly be called. This list is sorted, and
     /// trampolines for each of these signatures are required.
@@ -92,6 +81,14 @@ pub struct ModuleTranslation<'data> {
     /// `MemoryInitializer` type.
     pub data: Vec<Cow<'data, [u8]>>,
 
+    /// The desired alignment of `data` in the final data section of the object
+    /// file that we'll emit.
+    ///
+    /// Note that this is 1 by default but `MemoryInitialization::Static` might
+    /// switch this to a higher alignment to facilitate mmap-ing data from
+    /// an object file into a linear memory.
+    pub data_align: Option<u64>,
+
     /// Total size of all data pushed onto `data` so far.
     total_data: u32,
 
@@ -146,8 +143,8 @@ type Reader<'input> = gimli::EndianSlice<'input, gimli::LittleEndian>;
 #[allow(missing_docs)]
 pub struct NameSection<'a> {
     pub module_name: Option<&'a str>,
-    pub func_names: HashMap<u32, &'a str>,
-    pub locals_names: HashMap<u32, HashMap<u32, &'a str>>,
+    pub func_names: HashMap<FuncIndex, &'a str>,
+    pub locals_names: HashMap<FuncIndex, HashMap<u32, &'a str>>,
 }
 
 #[derive(Debug, Default)]
@@ -246,9 +243,12 @@ impl<'data> ModuleEnvironment<'data> {
                     .module
                     .functions
                     .iter()
-                    .filter_map(|(i, sig)| match self.result.module.defined_func_index(i) {
-                        Some(i) if !self.result.escaped_funcs.contains(&i) => None,
-                        _ => Some(*sig),
+                    .filter_map(|(_, func)| {
+                        if func.is_escaping() {
+                            Some(func.signature)
+                        } else {
+                            None
+                        }
                     })
                     .collect();
                 self.result.exported_signatures.sort_unstable();
@@ -390,7 +390,7 @@ impl<'data> ModuleEnvironment<'data> {
                     let sigindex = entry?;
                     let ty = TypeIndex::from_u32(sigindex);
                     let sig_index = self.result.module.types[ty].unwrap_function();
-                    self.result.module.functions.push(sig_index);
+                    self.result.module.push_function(sig_index);
                 }
             }
 
@@ -512,9 +512,6 @@ impl<'data> ModuleEnvironment<'data> {
             Payload::ElementSection(elements) => {
                 validator.element_section(&elements)?;
 
-                let cnt = usize::try_from(elements.get_count()).unwrap();
-                self.result.module.table_initializers.reserve_exact(cnt);
-
                 for (index, entry) in elements.into_iter().enumerate() {
                     let wasmparser::Element {
                         kind,
@@ -527,7 +524,7 @@ impl<'data> ModuleEnvironment<'data> {
                     // entries listed in this segment. Note that it's not
                     // possible to create anything other than a `ref.null
                     // extern` for externref segments, so those just get
-                    // translate to the reserved value of `FuncIndex`.
+                    // translated to the reserved value of `FuncIndex`.
                     let items_reader = items.get_items_reader()?;
                     let mut elements =
                         Vec::with_capacity(usize::try_from(items_reader.get_count()).unwrap());
@@ -576,15 +573,18 @@ impl<'data> ModuleEnvironment<'data> {
                                     )));
                                 }
                             };
-                            self.result
-                                .module
-                                .table_initializers
-                                .push(TableInitializer {
-                                    table_index,
-                                    base,
-                                    offset,
-                                    elements: elements.into(),
-                                });
+
+                            let table_segments = match &mut self.result.module.table_initialization
+                            {
+                                TableInitialization::Segments { segments } => segments,
+                                TableInitialization::FuncTable { .. } => unreachable!(),
+                            };
+                            table_segments.push(TableInitializer {
+                                table_index,
+                                base,
+                                offset,
+                                elements: elements.into(),
+                            });
                         }
 
                         ElementKind::Passive => {
@@ -616,7 +616,7 @@ impl<'data> ModuleEnvironment<'data> {
                 let func_index = FuncIndex::from_u32(func_index);
 
                 if self.tunables.generate_native_debuginfo {
-                    let sig_index = self.result.module.functions[func_index];
+                    let sig_index = self.result.module.functions[func_index].signature;
                     let sig = &self.types.wasm_signatures[sig_index];
                     let mut locals = Vec::new();
                     for pair in body.get_locals_reader()? {
@@ -874,7 +874,7 @@ impl<'data> ModuleEnvironment<'data> {
                                     self.result.module.num_imported_tables += 1;
                                 }
                                 EntityType::Function(sig) => {
-                                    self.result.module.functions.push(*sig);
+                                    self.result.module.push_function(*sig);
                                     self.result.module.num_imported_funcs += 1;
                                     self.result.debuginfo.wasm_file.imported_func_count += 1;
                                 }
@@ -1125,9 +1125,7 @@ and for re-adding support for interface types you can see this issue:
 
     fn push_type(&mut self, ty: EntityType) -> EntityIndex {
         match ty {
-            EntityType::Function(ty) => {
-                EntityIndex::Function(self.result.module.functions.push(ty))
-            }
+            EntityType::Function(ty) => EntityIndex::Function(self.result.module.push_function(ty)),
             EntityType::Table(ty) => {
                 let plan = TablePlan::for_table(ty, &self.tunables);
                 EntityIndex::Table(self.result.module.table_plans.push(plan))
@@ -1171,9 +1169,14 @@ and for re-adding support for interface types you can see this issue:
     }
 
     fn flag_func_escaped(&mut self, func: FuncIndex) {
-        if let Some(idx) = self.result.module.defined_func_index(func) {
-            self.result.escaped_funcs.insert(idx);
+        let ty = &mut self.result.module.functions[func];
+        // If this was already assigned an anyfunc index no need to re-assign it.
+        if ty.is_escaping() {
+            return;
         }
+        let index = self.result.module.num_escaped_funcs as u32;
+        ty.anyfunc = AnyfuncIndex::from_u32(index);
+        self.result.module.num_escaped_funcs += 1;
     }
 
     fn declare_type_func(&mut self, wasm: WasmFuncType) -> WasmResult<()> {
@@ -1291,18 +1294,17 @@ and for re-adding support for interface types you can see this issue:
                         if (index as usize) >= self.result.module.functions.len() {
                             continue;
                         }
+
+                        // Store the name unconditionally, regardless of
+                        // whether we're parsing debuginfo, since function
+                        // names are almost always present in the
+                        // final compilation artifact.
                         let index = FuncIndex::from_u32(index);
                         self.result
-                            .module
+                            .debuginfo
+                            .name_section
                             .func_names
-                            .insert(index, name.to_string());
-                        if self.tunables.generate_native_debuginfo {
-                            self.result
-                                .debuginfo
-                                .name_section
-                                .func_names
-                                .insert(index.as_u32(), name);
-                        }
+                            .insert(index, name);
                     }
                 }
                 wasmparser::Name::Module(module) => {
@@ -1332,7 +1334,7 @@ and for re-adding support for interface types you can see this issue:
                                 .debuginfo
                                 .name_section
                                 .locals_names
-                                .entry(f.indirect_index)
+                                .entry(FuncIndex::from_u32(f.indirect_index))
                                 .or_insert(HashMap::new())
                                 .insert(index, name);
                         }
